# Standard modules
import re
from pathlib import Path

# Pip modules
from numpy import array, ndarray


BLOCK_STATUS = [
    "?",    # 0 -> non_tried
    "+",    # 1 -> rescued
    "*",    # 2 -> non_trimmed
    "/",    # 3 -> non_scraped
    "-"     # 4 -> bad_sectors
]


RUN_STATUS = {
    "?": "copying non-tried blocks",
    "*": "trimming non-trimmed blocks",
    "/": "scraping non-scraped blocks",
    "-": "retrying bad sectors",
    "F": "filling the blocks specified",
    "G": "generating approximate mapfile",
    "+": "finished"
}


class MapEntry:
    def __init__(self, line_list):
        self.position = int(line_list[0], 0)
        self.size = int(line_list[1], 0)
        self.status = BLOCK_STATUS.index(line_list[2])


def parse_mapfile(file_path: Path, res_x: int, res_y: int) -> tuple[ndarray | None, dict]:
    # Not a (valid) file
    if not file_path or not isinstance(file_path, Path) or not file_path.exists():
        return None, dict({"msg": f"Invalid file path {file_path}"})
    # Array must be 2x2 or larger
    if not isinstance(res_x, int) or res_x < 2 or not isinstance(res_y, int) or res_y < 2:
        return None, dict({"msg": f"Invalid resolution {res_x}x{res_y}"})

    map_full_size = []
    general = dict.fromkeys(["version", "command", "direction", "action"], "")

    with open(file_path) as mapfile:
        is_first_line = True
        is_first_dataline = True
        for line in mapfile:
            line = line.strip()
            if is_first_line:
                is_first_line = False

                # Not a mapfile
                if not line.startswith("# Mapfile. Created by"):
                    return None, dict({"msg": f"Not a mapfile: {file_path}"})

            if line.startswith("# "):
                # heading comments
                if "forwards" in line:
                    general["direction"] = "forwards"
                if "backwards" in line:
                    general["direction"] = "backwards"
                if "Created by" in line:
                    general["version"] = line.split("Created by ")[1]
                if "Command line" in line:
                    general["command"] = line.split("Command line: ")[1]
            else:
                line = re.sub(' +', ' ', line).split(" ")
                if is_first_dataline:
                    is_first_dataline = False
                    general.update({
                        "cur_pos": int(line[0], 0),
                        "cur_status": RUN_STATUS[line[1]],
                        "cur_pass": int(line[2])
                    })
                else:
                    map_full_size.append(MapEntry(line))

    # Action line
    general["action"] = (f"{general['cur_status']}... Pass {general['cur_pass']}" +
                         (f" ({general['direction']})" if general["direction"] else ""))

    # After all lines are read, take last block and calculate divisor to reach targeted size
    map_full_size.sort(key=lambda x: x.position)
    res_len = (res_x * res_y)
    divisor = int((map_full_size[-1].position + map_full_size[-1].size) / res_len)

    map_plot_size = []
    size_0 = []
    for res_entry in map_full_size:
        size = int(res_entry.size / divisor)
        status = res_entry.status

        # Merge all size 0 to one entry
        if not size:
            size_0.append(status)
            continue

        # Add combined entry
        if size_0:
            size_0.append(status)
            map_plot_size.extend([max(size_0)] * len(size_0))
            size_0 = []
            size -= 1
            if not size:
                continue

        # Add normal entry
        map_plot_size.extend([status] * size)

    # if current position exists, mark it with black
    try:
        map_plot_size[int(general["cur_pos"] / divisor)] = 5
    except KeyError:
        pass

    # If block are missing or too many due to rounding errors, fill up with transparent
    while len(map_plot_size) < res_len:
        map_plot_size.append(6)
    while len(map_plot_size) > res_len:
        map_plot_size.pop()

    # Reshape 1d-array to 2d-array
    np_array_2d = array(map_plot_size).reshape((res_y, res_x))

    return np_array_2d, general
